/*
 * gps.c
 *
 * Created 14 Nov 2013
 * Author: Joel Heck
 * 
 * Adapted from code "gps.c" created by Brian on 30 Mar 2012 for a 4760
 * project
 *
 * Defines functions used for GPS data processing
 */

#include "main.h"
#include "gps.h"
#include "uart1.h"
#include <stdlib.h>
#include <string.h>

 #define UTC_OFFSET 5	// Hour difference between local and UTC time 
 						// Currently -5 hours (EDT)

// Command to turn on RMC and GGA output from GPS module
char* gps_strings= "$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28\r\n";
// 1 Hz update rate
char* gps_rate= "$PMTK220,1000*1F\r\n";

// Active GPS module
gps_data gps;

// Parses raw GPS data
static void parse_chunk(uint8_t* token);

/*
 * Receives a character containing raw GPS data from UART1 and stores it in
 * and array. Resets certain variables once all the data for a sentence has
 * been collected.
 * 
 * c: a character variable containing raw GPS data
 */
void receive_gps_uart(char c){
	//UDR0 = c;	// Print raw characters to PC (used for testing)
	
	// Make sure the sentence starts with '$' and is not empty
	if(c == 0 || (gps.rx_pos == 0 && c != '$'))
		return;

	gps.rx[gps.rx_pos] = c;

	// If character is a delimiter (comma) or null, advance state
	if(c == ',' || c == '\n'){
		gps.rx_status++;
		
		// If entire GPS sentence received, stop receiving data
		if(gps.rx_status == DONE){
			gps.data_collected = TRUE;
			gps.rx_status = GGA_TITLE;
			// Replace the last delimiter with a null character to help strtok
			gps.rx[gps.rx_pos] = 0;
			gps.rx_pos = 0;
			return;
		}
	}

	gps.rx_pos++;
}

/*
 * Initialize the GPS data collection variables and UART1
 * 
 * raw: a pointer to a Position structure to store parsed GPS data
 */
void init_gps(volatile Position* raw){
	gps.gps_status = GGA_TITLE;
	gps.data_collected = FALSE;
	gps.rx_pos = 0;
	gps.raw = raw;
	gps.rx_status = GGA_TITLE;

	// Initialize UART1 with the receive_gps_uart() function as the argument
	// so it will receive data from the UART1 receiver buffer
	uart1_init(receive_gps_uart);

	// Tell GPS to output RMC and GGA at 1 Hz
	uart1_send_buff(gps_strings,strlen(gps_strings));
	uart1_send_buff(gps_rate,strlen(gps_rate));
}

/*
 * Change received GPS time to integer seconds. The GPS time is given in 
 * hhmmss.ss format and is in UTC time standard. Discards fractional seconds
 * 
 * token: array of integers giving current UTC time
 *
 * return: given UTC time as integer seconds
 */
uint32_t gpstime_to_sec(const uint8_t* token){
	uint32_t ret = (uint32_t)((token[0]-48)*10+(token[1]-48))*3600;	// Hours
	ret += (uint32_t)((token[2]-48)*10+(token[3]-48))*60;	// Minutes
	ret += (uint32_t)((token[4]-48)*10+(token[5]-48));		// Seconds
	return ret;
}

/*
 * Change received GPS time to local time in hours and minutes. Store the 
 * local time in variable "raw". The GPS time is given in hhmmss.ss format and
 * is in UTC time standard. The time difference is simply a difference in 
 * hours. Discards seconds
 * 
 * token: array of integers giving each digit of current UTC time
 */
void gpstime_to_local(const uint8_t* token){
	uint8_t hours = (uint8_t)((token[0]-48)*10+(token[1]-48));	// Hours
	if (hours < UTC_OFFSET) 
		hours = 24 - (UTC_OFFSET-hours);
	else
		hours -= UTC_OFFSET;
	gps.raw->hour = hours;
	gps.raw->minute = (uint8_t)((token[2]-48)*10+(token[3]-48));	// Minutes
}

/*
 * Converts a receiver latitude or longitude position stored in a uint8_t
 * array to a double. Received latitude or longitude is given in ddmm.mmm
 * format. The returned format is dd.ddddd (decimal degrees).
 * 
 * token: array of integers giving each digit of latitude or longitude
 * dim: indicates whether input is a latitude or longitude 
 * 		0 = latitude, 1 = longitude
 *
 * return: given latitude or longitude in degrees as a double (dd.ddddd)
 */
double gps_pos(const uint8_t* token, uint8_t dim) {
	double ret = 0.0;

	if (dim == 0) {
		ret = (double)((token[0]-48)*10 + (token[1]-48));
		ret += (double)((token[2]-48)*10 + (token[3]-48) +
			(token[5]-48)*0.1 + (token[6]-48)*0.01 +
			(token[7]-48)*0.001)/60;

		return ret;
	} else {
		ret = (double)((token[0]-48)*100 + (token[1]-48)*10 + 
			(token[2]-48));
		ret += (double)((token[3]-48)*10 + (token[4]-48) + 
			(token[6]-48)*0.1 + (token[7]-48)*0.01 + 
			(token[8]-48)*0.001)/60;

		return ret;
		//return ret*-1;	// Negative for western hemisphere
	}
}


/*
 * Retrieves the degrees of a receiver latitude or longitude position stored
 * in a uint8_t array. Received latitude or longitude is given in ddmm.mmm
 * format. The returned value is the degrees as an integer.
 * 
 * token: array of integers giving each digit of latitude or longitude
 * dim: indicates whether input is a latitude or longitude 
 * 		0 = latitude, 1 = longitude
 *
 * return: integer degrees of given latitude or longitude
 */
uint8_t gps_deg(const uint8_t* token, uint8_t dim) {
	if (dim == 0) 
		return (uint8_t)((token[0]-48)*10 + (token[1]-48));
	else
		return (uint8_t)((token[0]-48)*100 + (token[1]-48)*10 + (token[2]-48));
}

/*
 * Retrieves the minutes of a receiver latitude or longitude position stored
 * in a uint8_t array. Received latitude or longitude is given in ddmm.mmm
 * format. The returned value is the minutes as an integer.
 * 
 * token: array of integers giving each digit of latitude or longitude
 * dim: indicates whether input is a latitude or longitude 
 * 		0 = latitude, 1 = longitude
 *
 * return: integer minutes of given latitude or longitude
 */
uint8_t gps_min(const uint8_t* token, uint8_t dim) {
	if (dim == 0) 
		return (uint8_t)((token[2]-48)*10 + (token[3]-48));
	else
		return (uint8_t)((token[3]-48)*10 + (token[4]-48));
}

/*
 * Retrieves the seconds of a receiver latitude or longitude position stored
 * in a uint8_t array. Received latitude or longitude is given in ddmm.mmm
 * format. The returned value is the seconds as an integer. Fractional seconds
 * are discarded
 * 
 * token: array of integers giving each digit of latitude or longitude
 * dim: indicates whether input is a latitude or longitude 
 * 		0 = latitude, 1 = longitude
 *
 * return: integer seconds of given latitude or longitude
 */
uint8_t gps_sec(const uint8_t* token, uint8_t dim) {
	if (dim == 0) 
		return (uint8_t)(((token[5]-48)*10 + (token[6]-48))*60/100);
	else
		return (uint8_t)(((token[6]-48)*10 + (token[7]-48))*60/100);
}

/*
 * Parses collected raw GPS data if all of it has been received. Returns the 
 * state of the GPS link. 
 * 
 * return: integer indicating the state of the GPS link
 * 		   -2 = bad fix, -1 = no received data, 0 = data received and parsed
 */
int8_t get_gps(){
	
	uint8_t* chunk;

	if(gps.data_collected){
		// Separate collected data using delimiters
		chunk = (uint8_t*) strtok((char*)gps.rx, ",\n");
		
		while(chunk != NULL){
			parse_chunk(chunk);
			chunk= (uint8_t*) strtok(NULL, ",\n");
		}

		gps.data_collected = FALSE;
   		
   		// Indicate that data should not be used if fix is bad
    	if(gps.fix != 1)
         	return -2;
   	    
		return 0;
	}

	return -1;
}

/*
 * Parse the input data based on the part of the GPS sentence it belongs to. 
 *
 * token: array of integers containing raw GPS data
 */
static void parse_chunk(uint8_t* token){
	switch(gps.gps_status){

		case GGA_TITLE:
			// Wait until start of GGA string to continue
			if(strcmp("$GPGGA",(const char*) token) == 0)
				gps.gps_status = GGA_UTC;
			break;

		// Store the UTC time
		case GGA_UTC:
			// Arbitrarily use this instead of time in RMC (both are same)
			gps.raw->gps_time = gpstime_to_sec((const uint8_t*)token);
			gpstime_to_local((const uint8_t*)token);
			gps.gps_status = GGA_LATITUDE;
			break;

		// Store the latitude
		case GGA_LATITUDE:
			gps.raw->latitude= gps_pos((const uint8_t*)token, 0);
			gps.gps_status= GGA_NS;
			break;

		// Latitude direction (determines sign of latitude)
		case GGA_NS:
			if(strcmp("S",(const char*) token) == 0){
				if (gps.raw->latitude > 0)
					gps.raw->latitude *= -1;
			}
			gps.gps_status= GGA_LONGITUDE;
			break;

		// Store the longitude
		case GGA_LONGITUDE:
			gps.raw->longitude = gps_pos((const uint8_t*)token, 1);
			gps.gps_status = GGA_EW;
			break;

		// Longitude direction (determines sign of longitude)
		case GGA_EW:
			if(strcmp("W",(const char*) token) == 0){
				if (gps.raw->longitude > 0)
					gps.raw->longitude *= -1;
			}
			gps.gps_status= GGA_FIX;
			break;

		// Store the GPS fix state
		case GGA_FIX:
			gps.fix = token[0]-48;
			gps.gps_status = GGA_SATELLITES;
			break;

		// Number of satellites in view (not used)
		case GGA_SATELLITES:
			gps.gps_status = GGA_HDOP;
			break;

		// Horizontal dilution of precision (not used)
		case GGA_HDOP:
			gps.gps_status = GGA_ALTITUDE;
			break;

		// Store altitude
		case GGA_ALTITUDE:
			gps.raw->altitude = atoi((const char*)token);
			gps.gps_status = GGA_ALTITUDE_UNITS;
			break;

		// Altitude units (not used)
		case GGA_ALTITUDE_UNITS:
			gps.gps_status = GGA_GEOID_HEIGHT;
			break;

		// Geoidal separation (not used)
		case GGA_GEOID_HEIGHT:
			gps.gps_status = GGA_GEOID_H_UNITS;
			break;

		// Units of geoidal separation (not used)
		case GGA_GEOID_H_UNITS:
			gps.gps_status = GGA_BLANK;
			break;

		// Age of differential GPS data (not used)
		case GGA_BLANK:
			gps.gps_status = GGA_CHECKSUM;
			break;

		// GPS checksum (not used)
		case GGA_CHECKSUM:
			gps.gps_status = RMC_TITLE;
			break;

		// Wait for RMC header
		case RMC_TITLE:
			if(strcmp("$GPRMC", (const char*)token) == 0)
				gps.gps_status = RMC_UTC;
			break;

		// UTC time (taken from GGA string)
		case RMC_UTC:
			gps.gps_status = RMC_STATUS;
			break;

		// Status indicator (not used)
		case RMC_STATUS:
			gps.gps_status = RMC_LATITUDE;
			break;

		// Latitude (taken from GGA string)
		case RMC_LATITUDE:
			gps.gps_status = RMC_NS;
			break;

		// Latitude direction (taken from GGA string)
		case RMC_NS:
			gps.gps_status = RMC_LONGITUDE;
			break;

		// Longitude (taken from GGA string)
		case RMC_LONGITUDE:
			gps.gps_status = RMC_EW;
			break;

		// Longitude direction (taken from GGA string)
		case RMC_EW:
			gps.gps_status = RMC_SPEED;
			break;

		// Speed in knots over ground
		case RMC_SPEED:
	    	// Convert GPS speed in knots to KPH
			gps.raw->speed = (int8_t)(atof((const char*)token)*1.852);
			gps.gps_status = RMC_HEADING;
			break;

		// True course in degrees
		case RMC_HEADING:
			gps.raw->heading = atoi((const char*)token);
			gps.gps_status = RMC_DATE;
			break;

		// Date in DDMMYY (not used)
		case RMC_DATE:
			gps.gps_status = RMC_MAGNETIC_VAR;
			break;

		// Magnetic variation in degrees (not used)
		case RMC_MAGNETIC_VAR:
			gps.gps_status = RMC_VAR_DIRECTION;
			break;

		// Magnetic variation direction (not used)
		case RMC_VAR_DIRECTION:
			gps.gps_status = RMC_CHECKSUM;
			break;

		// Checksum (not used)
		case RMC_CHECKSUM:
			gps.gps_status = GGA_TITLE;
			// Indicate that a full set of data has been collected
			//gps.data_collected = TRUE;
			break;

		case DONE:
			//should never get here
			break;
	}

}
